/*
 * Halt		Stop the system running.
 *		It re-enables CTRL-ALT-DEL, so that a hard reboot can
 *		be done. If called as reboot, it will reboot the system.
 *
 *		If the system is not in runlevel 0 or 6, halt will just
 *		execute a "shutdown -h" to halt the system, and reboot will
 *		execute an "shutdown -r". This is for compatibility with
 *		sysvinit 2.4.
 *
 * Usage:	halt [-n] [-w] [-d] [-f] [-p] [-r]
 *		-n: don't sync before halting the system
 *		-w: only write a wtmp reboot record and exit.
 *		-d: don't write a wtmp record.
 *		-f: force halt/reboot, don't call shutdown.
 *		-p: power down the system (if possible, otherwise halt)
 *		-r: reboot (act like "reboot")
 *
 *		Reboot and halt are both this program. Reboot
 *		is just a link to halt. Invoking the program
 *		as poweroff implies the -p option.
 *
 * Author:	Miquel van Smoorenburg, miquels@cistron.nl
 *
 * Version:	2.74,  24-Feb-1998
 *
 *		This file is part of the sysvinit suite,
 *		Copyright 1991-1998 Miquel van Smoorenburg.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/errno.h>
#include <sys/mount.h>
#include <stdio.h>
#include <stdlib.h>
#include <utmp.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
// #include <sys/times.h>
// #include <time.h>
#include <signal.h>
// #include <stdio.h>
// #include <getopt.h>
#include <termios.h>
// #include <linux/module.h>
#include <linux/version.h>
#include "reboot.h"

/* static char *Version = "@(#)halt  2.74  24-Feb-1998 miquels@cistron.nl"; */ /* unused */
char *progname;

#define KERNEL_MONITOR	1 /* If halt() puts you into the kernel monitor. */
#define RUNLVL_PICKY	0 /* Be picky about the runlevel */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,11)
#ifndef MNT_DETACH
#define MNT_DETACH 0x00000002
#endif
#define UMOUNT(x) umount2(x,MNT_DETACH)
#else
#define UMOUNT(x) umount(x)
#endif

extern int ifdown(void);
extern void write_wtmp(char *user, char *id, int pid, int type, char *line);

/*
 *	Send usage message.
 */
void usage()
{
	char *ustr[]={"usage: ",progname," [-n] [-w] [-d] [-f] [-i] [-p]\n"};
	int i;
	for(i=0;i<3;i++) write(2,ustr[i],strlen(ustr[i]));
	exit(1);
}

/*
 *	See if we were started directly from init.
 *	Get the runlevel from /var/run/utmp or the environment.
 */
int get_runlevel()
{
	struct utmp *ut;
	char *r;
#if RUNLVL_PICKY
	time_t boottime;
#endif

	/*
	 *	First see if we were started directly from init.
	 */
	if (getenv("INIT_VERSION") && (r = getenv("RUNLEVEL")) != NULL)
		return *r;

	/*
	 *	Hmm, failed - read runlevel from /var/run/utmp..
	 */
#if RUNLVL_PICKY
	/*
	 *	Get boottime from the kernel.
	 */
	time(&boottime);
	boottime -= (times(NULL) / HZ);
#endif

	/*
	 *	Find runlevel in utmp.
	 */
	setutent();
	while ((ut = getutent()) != NULL) {
#if RUNLVL_PICKY
		/*
		 *	Only accept value if it's from after boottime.
		 */
		if (ut->ut_type == RUN_LVL && ut->ut_time > boottime)
			return (ut->ut_pid & 255);
#else
		if (ut->ut_type == RUN_LVL)
			return (ut->ut_pid & 255);
#endif
	}
	endutent();

	/* This should not happen but warn the user! */
	{
	 char *msg[]={"WARNING: could not determine runlevel - doing soft ",progname,".\n"};
	 int i;
	 for(i=0;i<3;i++) write(2,msg[i],strlen(msg[i]));
	}

	return -1;
}

/*
 *	Switch to another runlevel.
 */
void do_shutdown(char *fl, char *tm)
{
	char *args[8];
	int i = 0;

	args[i++] = "shutdown";
	args[i++] = fl;
	if (tm) {
		args[i++] = "-t";
		args[i++] = tm;
	}
	args[i++] = "now";
	args[i++] = NULL;

	execv("/sbin/shutdown", args);
	execv("/etc/shutdown", args);
	execv("/bin/shutdown", args);

	perror("shutdown");
	exit(1);
}

/*
 *	Main program if init is called as "halt".
 *	Write a wtmp entry and reboot cq. halt.
 */
int halt_main(int argc, char **argv)
{
	int do_reboot = 0;
	int do_sync = 1;
	int do_wtmp = 1;
	int do_nothing = 0;
	int do_hard = 0;
	int do_ifdown = 0;
	int do_poweroff = 0;
	int c;
	char *tm = NULL;

	/*
	 *	Find out who we are
	 */
	if ((progname = strrchr(argv[0], '/')) != NULL)
		progname++;
	else
		progname = argv[0];

	if (geteuid() != 0) {
		char *msg[]={progname,": must be superuser.\n"};
		int i;
		for(i=0;i<2;i++) write(2,msg[i],strlen(msg[i]));
		exit(1);
	}

	if (!strcmp(progname, "reboot")) do_reboot = 1;
	if (!strcmp(progname, "poweroff")) do_poweroff = 1;

	/*
	 *	Get flags
	 */
	while((c = getopt(argc, argv, ":idfnprwt:")) != EOF) {
		switch(c) {
			case 'n':
				do_sync = 0;
				do_wtmp = 0;
				break;
			case 'w':
				do_nothing = 1;
				break;
			case 'd':
				do_wtmp = 0;
				break;
			case 'f':
				do_hard = 1;
				break;
			case 'i':
				do_ifdown = 1;
				break;
			case 'p':
				do_poweroff = 1;
				break;
			case 'r':
				do_reboot = 1;
				break;
			case 't':
				tm = optarg;
				break;
			default:
				usage();
		}
	 }
	if (argc != optind) usage();

	(void)chdir("/");

	if (!do_hard && !do_nothing) {
		/*
		 *	See if we are in runlevel 0 or 6.
		 */
		c = get_runlevel();
		if (c != '0' && c != '6')
			do_shutdown(do_reboot ? "-r" : "-h", tm);
	}

	/*
	 *	Record the fact that we're going down
	 */
	if (do_wtmp)
		write_wtmp("shutdown", "~~", 0, RUN_LVL, "~~");

	/*
	 *	Exit if all we wanted to do was write a wtmp record.
	 */
	if (do_nothing) exit(0);

	if (do_sync) {
		sync();
		sleep(2);
	}

	if (do_ifdown)
		(void)ifdown();

	/* KNOPPIX: Umount /KNOPPIX, unload/eject CD-Rom */
	/* All these are really syscalls */
	for(c=1;c<=3;c++) /* Attempt to umount this many times */
	 {
	  if(UMOUNT("/KNOPPIX")==0)
	   {
	    char *msg1="[1;36mPlease remove CD, close cdrom drive and hit return.[1;39m";
	    char *msg2="\r[K[1;34mShutdown complete.[0;39m";
	    /* missing in linux/module.h */
	    extern int delete_module(const char *);
	    usleep(250000); delete_module("cloop"); usleep(250000);
	    if(UMOUNT("/cdrom")!=EBUSY)
	     {write(2,msg1,strlen(msg1)); tcflush(1,TCIFLUSH); getchar();
	      write(2,msg2,strlen(msg2)); usleep(250000);}
	    break;
	   }
	  else
	   {
	    if(errno==EBUSY)
	     {
	      char *msg=c<3?"[1;31mCan't unmount CD yet...[0;39m\n":
	                "[1;36mCan't unmount CD, please reset.[0;39m\n";
	      write(2,msg,strlen(msg));
	     }
	    sleep(3);
	   }
	 }

	if (do_reboot) {
		init_reboot(BMAGIC_REBOOT);
	} else {
		/*
		 *	Turn on hard reboot, CTRL-ALT-DEL will reboot now
		 */
		init_reboot(BMAGIC_HARD);
		/*
		 *	Stop init; it is insensitive to the signals sent
		 *	by the kernel.
		 */
		kill(1, SIGTSTP);
		/*
		 *	Halt or poweroff.
		 */

		if (do_poweroff)
		{
			init_reboot(BMAGIC_POWEROFF);
			sleep(2);
		}
		/*
		 *	Fallthrough if failed.
		 */
		init_reboot(BMAGIC_HALT);
		sleep(2);
	}

	/*
	 *	If we return, we (c)ontinued from the kernel monitor.
	 */
	init_reboot(BMAGIC_SOFT);
	kill(1, SIGCONT);

	return(0);
}
